/*

Power_Router est un système qui permet d'utiliser l'excédent d'énergie autoproduit par 
l'allumage d'un appareil résistif (facteur de puissance proche de 1) ce qui évite l'injection 
au réseau public de distribution d'électricité.

Le principe de fonctionnement est le suivant :
- détection de phase entre courant et tension permet de savoir si on consomme ou bien on injecte
- en cas d'injection il se produit la mise en route progressive d'un dispositif d'absorption 
d'excédent de puissance 
- la mesure du courant permet d'ajuster au mieux le niveau d'absorption de cet excédent.
- Par ailleurs il est prévu une sortie temporisée de 30 secondes (paramétrable) lorsque le seuil 
d'injection est proche de 3W (paramétrable) permettant par exemple de couper l'injection d'une 
éolienne au profit de la charge de batteries. 

le programme prévoit :
- une sonde de tension : simple transfo 230V/5V Crête à Crête sur mi-tension (2.5V)
- une sonde de courant : 20A/25mA sur mi-tension (2.5V)
- un module de commande par triac
- un dispositif de détection de passage à zéro de la sinusoïde de tension secteur (par exemple 
l'optocoupleur H11AA1)
- la bibliothèque TimeOne.h à installer et disponible là : 
http://www.arduino.cc/playground/Code/Timer
- en option un afficheur LCD 1602 avec extension I2C

La gamme de puissances testée est va de 300 à 1000W

merci à Ryan McLaughlin <ryanjmclaughlin@gmail.com> pour avoir étudié et mis au point la partie 
commande du triac il y a quelques années et que j'ai repris dans ce programme :)
source : https://web.archive.org/web/20091212193047/http://www.arduino.cc:80/cgi-bin/yabb2/YaBB.pl?num=1230333861/15

_________________________________________________________________
|                                                               |
|       auteur : Philippe de Craene <dcphilippe@yahoo.fr        |
|           pour l' Association P'TITWATT                       |
_________________________________________________________________

Toute contribution en vue de l’amélioration de l’appareil est la bienvenue ! Il vous est juste
demandé de conserver mon nom et mon email dans l’entête du programme, et bien sûr de partager 
avec moi cette amélioration. Merci.

Récapitulatif des branchements :
 - broche A0 (analogique 0) => sonde de courant
 - broche A1 (analogique 1) => sonde de tension
 - broche A4 (analogique 4) => broche SDA sur LCD 
 - broche A5 (analogique 5) => broche SCL sur LCD
 - broche  2 (numérique 2)  => broche ZC (zero-cross) du module triac
 - broche  3 (numérique 3)  => LED indiquant le taux de pwm
 - broche 10 (numérique 10) => broche PWM (commande Triac) du module triac
 - brcohe 11 (numérique 11) => circuit de commande du délestage
 - broche 13 (numérique 13) => LED indiquant l'overflow

Chronologie des versions :
version 0.5 - 3 mai 2018     - boucle de décrémentation dim --
version 0.8 - 5 juil. 2018   - 1ère version fonctionnelle, pb du pic de courant du triac 
version 1   - 6 juil. 2018   - ajout de la bibliothèque EmonLib.h pour mesure du secteur
version 1.4 - 7 juil. 2018   - simplification des tests sur sPower et dim.
version 1.6 - 8 juil. 2018   - ajout LED d'overflow + optimisation des paramètres + seuilPoff
version 1.8 - 24 sept 2018   - ajout du pas variable sur dim avec dimstep
version 1.9 - 12 oct. 2018   - ajout d'une sortie temporisée de 5min à seuilPoff (25W) du seuil d'injection  
version 2.0 - 4 nov. 2018    - ajout d'un watchdog avec comptage de reset en EEPROM
version 2.2 - 7 nov. 2018    - seuilPtempo variable à part entière pour le délestage + correction coquille
version 2.3 - 16 dec 2018    - réaménagemet des messages console pour gagner du temps
version 2.4 - 12 jan 2019    - ajout d'un afficheur LCD 1602 avec extension I2C
version 3.2 - 17 jan 2019    - gain en performances en contournant EmonLib.h
version 3.3 - 22 fev 2019    - abandon de seuilPoff : arrêt en cas de chutte brusque d'injection 

*/

#include <EEPROM.h>
#include <avr/wdt.h>    // documentation : https://tushev.org/articles/arduino/5/arduino-and-watchdog-timer
#include <TimerOne.h>   // librairie à installer depuis http://www.arduino.cc/playground/Code/Timer1
#include <LiquidCrystal_I2C.h> // librairie à installer depuis https://app.box.com/s/czde88f5b9vpulhf8z56
// Attention ! Renomer la librairie d'origine de l'IDE LiquidCrystal.h et LiquidCrystal.cpp

// variables de calibration :

bool CALIBRATION = false; // pour définir les valeurs de Vcalibration et Icalibration
bool VERBOSE = false;     // pour le debugging VERBOSE = true affiche les mesures à la console 
                          // cependant au détriment des performances

// Calibration des mesures des sondes pour avoir des valeurs correctes :
// en premier : Vcalibration doit être défini pour obtenir les 230V du secteur
// ensuite : Icalibration en comparant l'affichage de la console avec celui d’un wattmètre 
// pour s'approcher de la réalité. En l'absence de wattmètre effectuer le test avec une 
// charge résistive connue, par exemple une ampoule halogène neuve.
// REMARQUE : les variables de type float exigent un nombre à virgule

float Vcalibration     = 0.97;  // pour obtenir 230V
float Icalibration     = 40.6;  // le courant est mesuré en mA (milliampères)
float phasecalibration = 1.7;   // valeur empirique compensant le déphasage lié aux capteurs...
byte totalCount        = 20;    // nombre de demie-périodes étudiées pour la mesure

// Valeurs en mW (milliwatts) qui déterminent les seuils : 

int seuilP             = 3000;  // l'hystérésis d'asservissement : 3000 = 3W => hystérésis à 6W
int valtempo           = 30;    // temporisation du délestage temporisé : 30 secondes
int seuilPtempo        = 3000;  // seuil de puissance pour démarrage du délestage  
bool etat_tempo_repos  = HIGH;  // état de la sortie temporisée au repos : HIGH pour actif

// Valeur du coefficient de réaction de l'asservissement avec le dimensionnement de dimstep :
// Incrémente dimstep par pas issue du rapport entre la puissance à dissiper et coefdeReaction
// compromis entre rapidité de réaction à l'injection et instabilité :
// trop petit = risque l'oscillation, trop grand = plus lent
// détermination de la valeur : coefdeReaction ~ (puissance de la charge)/4 en Watts

unsigned int coefdeReaction  = 90; 

// détermination des entrées / sorties :

const byte triac_pin         = 10;    // sortie numérique de commande de charge résistive
const byte tempo_pin         = 11;    // sortie temporisée à 5min à 20W du seuil 
const byte triacLED          = 3;     // sortie numérique pour la LED de test du triac
const byte limiteLED         = 13;    // sortie numérique pour la LED d'overflow
const byte voltageSensorPin  = 1;     // détecteur de tension entrée analogique 1
const byte currentSensorPin  = 0;     // détecteur de courant entrée analogique 0
const byte zeroCrossPin      = 2;     // détecteur de phase entrée digitale 2

// variables de gestion des interruptions (zero-crossing) :
 
byte dimmax = 128;              // valeur max de dim pour inhiber le triac
byte dim = dimmax;              // Dimming level (0-128)  0 = on, 128 = 0ff 
char periodStep = 75;           // détermine la période du timer (65 pour 60Hz, 78 pour 50Hz, en µs)
                                // suivant la formule (500000/AC_freq)/NumSteps = periodStep 
                                // 78*128=10ms=1/2 période 50Hz mais aux tests 76 marche mieux
volatile int i = 0;             // Variable to use as a counter
volatile bool zero_cross = false;       // indicateur de zero-cross pour commander le triac
volatile bool zero_cross_flag = false;  // indicateur de zero-cross pour calcul des puissances électriques

// variables de calcul des grandeurs électriques :

int lectureV, memo_lectureV, lectureI;   // tensions et courants en bits (0 à 1023 bits)
float rPower, V, I, sqV, sumV = 0, sqI, sumI = 0, instP, sumP = 0;             
byte zero_crossCount = 0;       // compteur de demie-périodes
    

// autres variables !

int dimstep;                    // valeur de l'écart de dim 
unsigned int decompte;          // décompte de la temporisation du délestage
unsigned int memo_temps = 0;   
bool tempoON = false;           // état du délestage

// Déclaration du LCD en mode I2C :

// toutes les infos ici : http://arduino-info.wikispaces.com/LCD-Blue-I2C
// Set the pins on the I2C chip used for LCD connections:  
//                   addr, en,rw,rs,d4,d5,d6,d7,bl,blpol  
LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);   
// => connexion des 2 broches I2C sur l'Arduino Uno R3 : SDA en A4, SCL en A5


//
// SETUP
//_____________________________________________________________________________________________

void setup() {                  // Begin setup
 pinMode(triac_pin, OUTPUT);    // Set the Triac pin as output
 pinMode(tempo_pin, OUTPUT);
 pinMode(triacLED,  OUTPUT);    // Set the LED pin as output
 pinMode(limiteLED, OUTPUT);     

 attachInterrupt(digitalPinToInterrupt(zeroCrossPin), zero_cross_detect, RISING);  
// à chaque changement d'état de zeroCrossPin la fonction 'zero_cross_detect' est exécutée
// quelque soit l'endroit où se trouvait l'exécution du programme : définition d'une interruption
// documentation : https://www.arduino.cc/reference/en/language/functions/external-interrupts/attachinterrupt/

 Timer1.initialize(periodStep);                    // initialisation de la librairie TimerOne
 Timer1.attachInterrupt(dim_check, periodStep);  
// autre mode d'interruption avec la librairie TimerOne.h : à chaque periodStep écoulé dont la 
// valeur est définie en microsecondes la fonction 'dim_check' est lancée : cette interruption 
// nous assure d'actionner le triac au bon moment
                                      
// Utilisation de EEPROM pour récupérer et mettre à jour le nombre de redémarrage du programme
// reconstruction du nombre redemarage, pour rappel EEPROM ne gère que le type unsigned char 
 unsigned char redemarrage_high = EEPROM.read(0);  // récupère la 1ère moitié du nombre
 unsigned char redemarrage_low  = EEPROM.read(1);  // récupère la 2ème moitié du nombre
 unsigned int redemarrage = (redemarrage_high << 8) + redemarrage_low; 
 redemarrage++;
 EEPROM.update(0, highByte(redemarrage));
 EEPROM.update(1, lowByte(redemarrage));

// initialisation du LCD
 lcd.begin(16,2);                // initialize the lcd for 16 chars 2 lines
 lcd.clear();                    
 lcd.setCursor(0, 0);            
 lcd.print("PTIWATT BONJOUR");  
 lcd.setCursor(0, 1);            
 lcd.print("   **********");  

// initialisation de la console
 Serial.begin(250000);
 Serial.println ();
 Serial.print("A NOTER : ");
 Serial.print(redemarrage);
 Serial.println(" ème redémarrage du programme");
 Serial.println(); 
 Serial.println("Prêt à démarrer ...");
 Serial.println ();
 delay(500); 
 if( VERBOSE == true ) Serial.print("  Pu (W) || dimstep |  dim ||  sortie temporisée");
 else Serial.println("C'est parti !"); 
 Serial.println();

 wdt_enable(WDTO_500MS);         // watchdog = reset après inaction supérieure à 500ms 
   
}                // End setup


//
// ZERO CROSS DETECT : gestion du passage à zéro par interruption
//____________________________________________________________________________________________

void zero_cross_detect() {   // fonction appelée à chaque passage à zéro de la tension secteur
   zero_cross_flag = true;   // témoin pour démarrer le calcul de la puissance
   zero_cross = true;        // témoin pour la commande du triac
}                                


//
// DIM CHECK : gestion du triac par interruption
//____________________________________________________________________________________________

void dim_check() {                        // Function will fire the triac at the proper time
 if(zero_cross == true && dim < dimmax) { // First check to make sure the zero-cross has 
                                          // happened else do nothing
     if(i>dim) {            // i est un compteur qui détermine le retard au fire. plus dim 
                            // est élevé, plus de temps prendra le compteur i et plus tard
       digitalWrite(triac_pin, HIGH);     // se fera le fire du triac 
       delayMicroseconds(50);             // Pause briefly to ensure the triac turned on
       digitalWrite(triac_pin, LOW);      // Turn off the Triac gate, le triac reste conducteur
       i = 0;                             // Reset the accumilator
       zero_cross = false;
     } 
    else { i++; }           // If the dimming value has not been reached, incriment our counter
 }                          // End zero_cross check
}                           // End dim_check function


//
// LOOP : programme principal (qui tourne en boucle)
//____________________________________________________________________________________________

void loop() {                  // Main Loop

// 1ère partie : calcul de la puissance réelle relevée par les capteurs avec rPower
//____________________________________________________________________________________________
//
  
  unsigned int numberOfSamples = 0;
  sumV = 0;
  sumI = 0;
  sumP = 0;
  
// à chaque passage à zéro de la tension du secteur réinitialisation périodique du compteur de 
// passage à zéro zero_crossCount lorque le nombre de cycles de mesures totalCount est atteint 

 if( zero_crossCount >= totalCount ) { zero_crossCount = 0; }

// processus de relevé de mesures durant le nombre défini de demi-périodes 
  while( zero_crossCount < totalCount ) {
    if( zero_cross_flag == true ) {        // incrémentation du compteur de demi-périodes secteur
      zero_cross_flag = false;
      zero_crossCount++; 
    }  
    numberOfSamples++;                         // comptage du nombre d'itérations
    memo_lectureV = lectureV;                  // mémorisation du relevé précédent
    lectureV = analogRead(voltageSensorPin);   // mesure de V en bits - 0V = bit 512
    lectureI = analogRead(currentSensorPin);   // mesure de I en bits - 0A = bit 512

// calcul des valeurs efficaces de tension et courant
    if( CALIBRATION == true ) {                    // ne sert que pour définir la calibration de V et I
      sqV= (lectureV -512.0) * (lectureV -512.0);       // -512 comme offset pour ramener 0V à bit 0
      sumV += sqV;               
      sqI = (lectureI -512.0) * (lectureI -512.0);
	    sumI += sqI;
    }    // fin de test sur VERBOSE
// calcul de la puissance instantanée 
    instP = ((memo_lectureV -512.0) + phasecalibration * ((lectureV -512.0) - (memo_lectureV -512))) * (lectureI -512.0); 
    sumP +=instP;  
 }      // fin de while sur zero_crossCount

// mémorisation des valeurs électriques
  if( numberOfSamples > 0 ) {
    if( CALIBRATION == true ) { 
      V = Vcalibration * sqrt(sumV / numberOfSamples);
      I = Icalibration * sqrt(sumI / numberOfSamples);
    }
    rPower = Vcalibration * Icalibration * sumP / numberOfSamples;
  }
	
// 2ème partie : calcul du taux de puissance à délester pour ne pas injecter avec dim et dimstep
//               + détection si nécessité de délestage avec tempoON
//____________________________________________________________________________________________
// 

// calcul de dimstep : le pas d'incrémentation de dim qui dépend de la puissance en jeu
  if( rPower > 0 ) { dimstep = (rPower/1000)/coefdeReaction + 1; } 
  else { dimstep = 1 - (rPower/1000)/coefdeReaction; }
  
  if( rPower > seuilP ) {     // l'injection augmente, on diminue le délai d'allumage du triac
    if( dim > dimstep )  dim -= dimstep; else  dim = 0;
  } 
  else if( rPower < -seuilP ) {                 // moins de prod : on baisse la charge
    if( dim + dimstep < dimmax ) dim += dimstep;  else  dim = dimmax; 
  }

  if(dim < 1) { digitalWrite(limiteLED, HIGH); }  // led témoin de surcharge
  else { digitalWrite(limiteLED, LOW); }
  analogWrite(triacLED, dimmax-dim);              // Write the value to the LED for testing

// Sortie temporisée quand le seuil seuilPtempo est atteint
  unsigned int temps_actuel = millis()/1000;      // mesure du temps en secondes
 
  if( rPower > -seuilPtempo) { 
    tempoON = true;                               // détection seuil atteint
    decompte = temps_actuel;                      // initialisation du compteur
  }
   
  if( tempoON = true ) {
    if( temps_actuel - decompte < valtempo) { 
      digitalWrite(tempo_pin, !etat_tempo_repos); // sortie temporisée active 
    }
    else {
      digitalWrite(tempo_pin, etat_tempo_repos);
      tempoON = false;                            // fin de temporisation
    }
  }   // fin de test sur tompoON

// affichage de ce qui se passe toutes les 2 secondes car ça bouffe du temps....
  if( temps_actuel >= memo_temps +2 ) {
    memo_temps = temps_actuel;
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("P= ");
    lcd.print(String(-rPower/1000,0));   
    lcd.print("w");
    lcd.setCursor(10, 0);
    lcd.print("T= ");
    lcd.print( map(dim, 0, dimmax, 99, 0) );
    lcd.print("%");
    lcd.setCursor(0, 1);   
    lcd.print("DELESTAGE ");              
    if( tempoON == true ) {  
      lcd.print(temps_actuel - decompte);
      lcd.print("s    ");
    }
    else { lcd.print("ARRETE"); }
  }  // fin tempo de 2 secondes

  if( CALIBRATION == true ) {
	  Serial.print(V);
	  Serial.print("  |  ");
    Serial.print(I/1000);
    Serial.print("  |  ");
    Serial.print(rPower/1000);
    Serial.println();
  }
  if( VERBOSE == true ) {
    Serial.print(rPower/1000);
    Serial.print("  ||     ");
    Serial.print(dimstep);
    Serial.print("  |  ");
    Serial.print(dim);
    Serial.print(" ||  ");
    Serial.print(" état tempo : ");
    Serial.print(tempoON);
    Serial.print(" décompte : ");
    Serial.println(temps_actuel - decompte);
  }
  else { delay(1); }           // obligatoire pour la stabilité

 wdt_reset();                  // raz de la tempo du watchdog
 
}                              // fin de Main Loop
